@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*  GMime
- *  Copyright (C) 2000-2008 Jeffrey Stedfast
+ *  Copyright (C) 2000-2010 Jeffrey Stedfast
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public License
@@ -245,7 +245,7 @@ g_mime_filter_basic_new (GMimeContentEncoding encoding, gboolean encode)
 {
 	GMimeFilterBasic *basic;
 	
-	basic = g_object_new (GMIME_TYPE_FILTER_BASIC, NULL);
+	basic = g_object_newv (GMIME_TYPE_FILTER_BASIC, 0, NULL);
 	
 	if (encode)
 		g_mime_encoding_init_encode (&basic->encoder, encoding);
