@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*  GMime
- *  Copyright (C) 2000-2007 Jeffrey Stedfast
+ *  Copyright (C) 2000-2010 Jeffrey Stedfast
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public License
@@ -130,9 +130,9 @@ filter_filter (GMimeFilter *filter, char *in, size_t len, size_t prespace,
 	GMimeFilterFrom *from = (GMimeFilterFrom *) filter;
 	struct fromnode *head = NULL, *tail = (struct fromnode *) &head, *node;
 	register char *inptr, *inend;
-	unsigned int left;
 	int fromcount = 0;
 	char *outptr;
+	size_t left;
 	
 	inptr = in;
 	inend = inptr + len;
@@ -146,7 +146,7 @@ filter_filter (GMimeFilter *filter, char *in, size_t len, size_t prespace,
 		}
 		
 		if (c == '\n' || !from->midline) {
-			left = inend - inptr;
+			left = (size_t) (inend - inptr);
 			if (left > 0) {
 				from->midline = TRUE;
 				if (left < 5) {
@@ -202,14 +202,14 @@ filter_filter (GMimeFilter *filter, char *in, size_t len, size_t prespace,
 			node = node->next;
 		}
 		
-		memcpy (outptr, inptr, (unsigned) (inend - inptr));
+		memcpy (outptr, inptr, (size_t) (inend - inptr));
 		outptr += inend - inptr;
 		*out = filter->outbuf;
-		*outlen = outptr - filter->outbuf;
+		*outlen = (size_t) (outptr - filter->outbuf);
 		*outprespace = filter->outbuf - filter->outreal;
 	} else {
 		*out = in;
-		*outlen = inend - in;
+		*outlen = (size_t) (inend - in);
 		*outprespace = prespace;
 	}
 }
@@ -242,17 +242,17 @@ filter_reset (GMimeFilter *filter)
  * will be escaped to ">From ".
  *
  * Note: If you plan on using a from-filter in mode ARMOR, you should
- * remember to also use a #GMimeFilterBasic filter with mode
- * #GMIME_FILTER_BASIC_QP_ENC.
+ * remember to also use a #GMimeFilterBasic filter with an encoding of
+ * #GMIME_CONTENT_ENCODING_QUOTEDPRINTABLE.
  *
- * Returns a new from filter with mode @mode.
+ * Returns: a new from filter with mode @mode.
  **/
 GMimeFilter *
 g_mime_filter_from_new (GMimeFilterFromMode mode)
 {
 	GMimeFilterFrom *new;
 	
-	new = g_object_new (GMIME_TYPE_FILTER_FROM, NULL);
+	new = g_object_newv (GMIME_TYPE_FILTER_FROM, 0, NULL);
 	new->midline = FALSE;
 	switch (mode) {
 	case GMIME_FILTER_FROM_MODE_ARMOR:
