@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*  GMime
- *  Copyright (C) 2000-2008 Jeffrey Stedfast
+ *  Copyright (C) 2000-2010 Jeffrey Stedfast
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public License
@@ -42,7 +42,7 @@
 
 /**
  * SECTION: gmime-multipart-encrypted
- * @title: GMimeMultpartEncrypted
+ * @title: GMimeMultipartEncrypted
  * @short_description: Encrypted MIME multiparts
  * @see_also: #GMimeMultipart
  *
@@ -119,7 +119,6 @@ g_mime_multipart_encrypted_class_init (GMimeMultipartEncryptedClass *klass)
 static void
 g_mime_multipart_encrypted_init (GMimeMultipartEncrypted *mpe, GMimeMultipartEncryptedClass *klass)
 {
-	mpe->protocol = NULL;
 	mpe->decrypted = NULL;
 	mpe->validity = NULL;
 }
@@ -129,8 +128,6 @@ g_mime_multipart_encrypted_finalize (GObject *object)
 {
 	GMimeMultipartEncrypted *mpe = (GMimeMultipartEncrypted *) object;
 	
-	g_free (mpe->protocol);
-	
 	if (mpe->decrypted)
 		g_object_unref (mpe->decrypted);
 	
@@ -173,13 +170,6 @@ multipart_encrypted_remove_header (GMimeObject *object, const char *header)
 static void
 multipart_encrypted_set_content_type (GMimeObject *object, GMimeContentType *content_type)
 {
-	GMimeMultipartEncrypted *mpe = (GMimeMultipartEncrypted *) object;
-	const char *protocol;
-	
-	protocol = g_mime_content_type_get_parameter (content_type, "protocol");
-	g_free (mpe->protocol);
-	mpe->protocol = g_strdup (protocol);
-	
 	GMIME_OBJECT_CLASS (parent_class)->set_content_type (object, content_type);
 }
 
@@ -209,7 +199,7 @@ g_mime_multipart_encrypted_new (void)
 	GMimeMultipartEncrypted *multipart;
 	GMimeContentType *content_type;
 	
-	multipart = g_object_new (GMIME_TYPE_MULTIPART_ENCRYPTED, NULL);
+	multipart = g_object_newv (GMIME_TYPE_MULTIPART_ENCRYPTED, 0, NULL);
 	
 	content_type = g_mime_content_type_new ("multipart", "encrypted");
 	g_mime_object_set_content_type (GMIME_OBJECT (multipart), content_type);
@@ -223,7 +213,7 @@ g_mime_multipart_encrypted_new (void)
  * g_mime_multipart_encrypted_encrypt:
  * @mpe: multipart/encrypted object
  * @content: MIME part to encrypt
- * @ctx: encryption cipher context
+ * @ctx: encryption context
  * @sign: %TRUE if the content should also be signed or %FALSE otherwise
  * @userid: user id to use for signing (only used if @sign is %TRUE)
  * @recipients: an array of recipients to encrypt to
@@ -295,7 +285,6 @@ g_mime_multipart_encrypted_encrypt (GMimeMultipartEncrypted *mpe, GMimeObject *c
 	g_object_unref (wrapper);
 	g_object_unref (stream);
 	
-	mpe->protocol = g_strdup (ctx->encrypt_protocol);
 	mpe->decrypted = content;
 	g_object_ref (content);
 	
@@ -315,7 +304,8 @@ g_mime_multipart_encrypted_encrypt (GMimeMultipartEncrypted *mpe, GMimeObject *c
 	g_object_unref (version_part);
 	
 	/* set the content-type params for this multipart/encrypted part */
-	g_mime_object_set_content_type_parameter (GMIME_OBJECT (mpe), "protocol", mpe->protocol);
+	g_mime_object_set_content_type_parameter (GMIME_OBJECT (mpe), "protocol",
+						  ctx->encrypt_protocol);
 	g_mime_multipart_set_boundary (GMIME_MULTIPART (mpe), NULL);
 	
 	return 0;
@@ -325,19 +315,23 @@ g_mime_multipart_encrypted_encrypt (GMimeMultipartEncrypted *mpe, GMimeObject *c
 /**
  * g_mime_multipart_encrypted_decrypt:
  * @mpe: multipart/encrypted object
- * @ctx: decryption cipher context
+ * @ctx: decryption context
  * @err: a #GError
  *
  * Attempts to decrypt the encrypted MIME part contained within the
- * multipart/encrypted object @mpe using the @ctx decryption context.
+ * multipart/encrypted object @mpe using the @ctx decryption
+ * context.
  *
- * If @validity is non-NULL, then on a successful decrypt operation,
- * it will be updated to point to a newly-allocated
- * #GMimeSignatureValidity with signature status information.
+ * For information regarding the signature status (in cases where the
+ * MIME part was both encrypted and signed), you may wish to check the
+ * results of g_mime_multipart_encrypted_get_signature_validity().
  *
- * Returns: the decrypted MIME part on success or %NULL on fail. If the
- * decryption fails, an exception will be set on @err to provide
+ * Returns: the decrypted MIME part on success or %NULL on fail. If
+ * the decryption fails, an exception will be set on @err to provide
  * information as to why the failure occured.
+ *
+ * Note: On success, the decrypted MIME part is cached internally and
+ * so the returned #GMimeObject must NOT be unreffed.
  **/
 GMimeObject *
 g_mime_multipart_encrypted_decrypt (GMimeMultipartEncrypted *mpe, GMimeCipherContext *ctx,
@@ -369,7 +363,8 @@ g_mime_multipart_encrypted_decrypt (GMimeMultipartEncrypted *mpe, GMimeCipherCon
 		/* make sure the protocol matches the cipher encrypt protocol */
 		if (g_ascii_strcasecmp (ctx->encrypt_protocol, protocol) != 0) {
 			g_set_error (err, GMIME_ERROR, GMIME_ERROR_PROTOCOL_ERROR,
-				     "Failed to decrypt MIME part: protocol error");
+				     "Cannot decrypt multipart/encrypted part: unsupported encryption protocol '%s'.",
+				     protocol);
 			
 			return NULL;
 		}
@@ -383,8 +378,8 @@ g_mime_multipart_encrypted_decrypt (GMimeMultipartEncrypted *mpe, GMimeCipherCon
 	/* make sure the protocol matches the version part's content-type */
 	content_type = g_mime_content_type_to_string (version->content_type);
 	if (g_ascii_strcasecmp (content_type, protocol) != 0) {
-		g_set_error (err, GMIME_ERROR, GMIME_ERROR_PROTOCOL_ERROR,
-			     "Failed to decrypt MIME part: protocol error");
+		g_set_error (err, GMIME_ERROR, GMIME_ERROR_PROTOCOL_ERROR, "%s",
+			     "Cannot decrypt multipart/encrypted part: content-type does not match protocol.");
 		
 		g_free (content_type);
 		
@@ -396,7 +391,9 @@ g_mime_multipart_encrypted_decrypt (GMimeMultipartEncrypted *mpe, GMimeCipherCon
 	encrypted = g_mime_multipart_get_part (GMIME_MULTIPART (mpe), GMIME_MULTIPART_ENCRYPTED_CONTENT);
 	mime_type = g_mime_object_get_content_type (encrypted);
 	if (!g_mime_content_type_is_type (mime_type, "application", "octet-stream")) {
-		g_object_unref (encrypted);
+		g_set_error (err, GMIME_ERROR, GMIME_ERROR_PROTOCOL_ERROR, "%s",
+			     "Cannot decrypt multipart/encrypted part: unexpected content type");
+		
 		return NULL;
 	}
 	
@@ -431,8 +428,8 @@ g_mime_multipart_encrypted_decrypt (GMimeMultipartEncrypted *mpe, GMimeCipherCon
 	g_object_unref (parser);
 	
 	if (!decrypted) {
-		g_set_error (err, GMIME_ERROR, GMIME_ERROR_PARSE_ERROR,
-			     "Failed to decrypt MIME part: parse error");
+		g_set_error (err, GMIME_ERROR, GMIME_ERROR_PARSE_ERROR, "%s",
+			     "Cannot decrypt multipart/encrypted part: failed to parse decrypted content");
 		
 		g_mime_signature_validity_free (sv);
 		
