@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*  GMime
- *  Copyright (C) 2000-2008 Jeffrey Stedfast
+ *  Copyright (C) 2000-2010 Jeffrey Stedfast
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public License
@@ -26,16 +26,22 @@
 #include <stdlib.h>
 #include <string.h>
 #include <sys/types.h>
-#include <regex.h>
 
 #include "gmime-parser.h"
 
 #include "gmime-table-private.h"
-#include "gmime-stream-mem.h"
 #include "gmime-message-part.h"
+#include "gmime-parse-utils.h"
+#include "gmime-stream-mem.h"
 #include "gmime-multipart.h"
 #include "gmime-part.h"
 
+#if GLIB_MAJOR_VERSION > 2 || (GLIB_MAJOR_VERSION == 2 && GLIB_MINOR_VERSION >= 14)
+#define HAVE_GLIB_REGEX
+#elif defined (HAVE_REGEX_H)
+#include <regex.h>
+#endif
+
 #ifdef ENABLE_WARNINGS
 #define w(x) x
 #else
@@ -56,8 +62,22 @@
  * into multiple #GMimeMessage objects.
  **/
 
+typedef struct _boundary_stack {
+	struct _boundary_stack *parent;
+	char *boundary;
+	size_t boundarylen;
+	size_t boundarylenfinal;
+	size_t boundarylenmax;
+	gint64 content_end;
+} BoundaryStack;
 
-typedef struct _ContentType {
+typedef struct _header_raw {
+	struct _header_raw *next;
+	char *name, *value;
+	gint64 offset;
+} HeaderRaw;
+
+typedef struct _content_type {
 	char *type, *subtype;
 	gboolean exists;
 } ContentType;
@@ -113,10 +133,15 @@ struct _GMimeParserPrivate {
 	gint64 from_offset;
 	GByteArray *from_line;
 	
-	regex_t header_regex;
 	GMimeParserHeaderRegexFunc header_cb;
 	gpointer user_data;
 	
+#if defined (HAVE_GLIB_REGEX)
+	GRegex *regex;
+#elif defined (HAVE_REGEX_H)
+	regex_t regex;
+#endif
+	
 	/* header buffer */
 	char *headerbuf;
 	char *headerptr;
@@ -128,6 +153,10 @@ struct _GMimeParserPrivate {
 	size_t rawleft;
 	
 	/* current message headerblock offsets */
+	gint64 message_headers_begin;
+	gint64 message_headers_end;
+	
+	/* current mime-part headerblock offsets */
 	gint64 headers_begin;
 	gint64 headers_end;
 	
@@ -144,18 +173,9 @@ struct _GMimeParserPrivate {
 	unsigned short int persist_stream:1;
 	unsigned short int respect_content_length:1;
 	
-	struct _header_raw *headers;
+	HeaderRaw *headers;
 	
-	struct _boundary_stack *bounds;
-};
-
-struct _boundary_stack {
-	struct _boundary_stack *parent;
-	char *boundary;
-	size_t boundarylen;
-	size_t boundarylenfinal;
-	size_t boundarylenmax;
-	gint64 content_end;
+	BoundaryStack *bounds;
 };
 
 
@@ -163,12 +183,12 @@ static void
 parser_push_boundary (GMimeParser *parser, const char *boundary)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
-	struct _boundary_stack *s;
+	BoundaryStack *s;
 	size_t max;
 	
 	max = priv->bounds ? priv->bounds->boundarylenmax : 0;
 	
-	s = g_new (struct _boundary_stack, 1);
+	s = g_slice_new (BoundaryStack);
 	s->parent = priv->bounds;
 	priv->bounds = s;
 	
@@ -191,7 +211,7 @@ static void
 parser_pop_boundary (GMimeParser *parser)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
-	struct _boundary_stack *s;
+	BoundaryStack *s;
 	
 	if (!priv->bounds) {
 		d(g_warning ("boundary stack underflow"));
@@ -202,47 +222,42 @@ parser_pop_boundary (GMimeParser *parser)
 	priv->bounds = priv->bounds->parent;
 	
 	g_free (s->boundary);
-	g_free (s);
+	
+	g_slice_free (BoundaryStack, s);
 }
 
-struct _header_raw {
-	struct _header_raw *next;
-	char *name;
-	char *value;
-	gint64 offset;
-};
-
 static const char *
-header_raw_find (struct _header_raw *headers, const char *name, gint64 *offset)
+header_raw_find (HeaderRaw *headers, const char *name, gint64 *offset)
 {
-	struct _header_raw *h;
+	HeaderRaw *header = headers;
 	
-	h = headers;
-	while (h) {
-		if (!g_ascii_strcasecmp (h->name, name)) {
+	while (header) {
+		if (!g_ascii_strcasecmp (header->name, name)) {
 			if (offset)
-				*offset = h->offset;
-			return h->value;
+				*offset = header->offset;
+			return header->value;
 		}
 		
-		h = h->next;
+		header = header->next;
 	}
 	
 	return NULL;
 }
 
 static void
-header_raw_clear (struct _header_raw **headers)
+header_raw_clear (HeaderRaw **headers)
 {
-	struct _header_raw *h, *n;
+	HeaderRaw *header, *next;
 	
-	h = *headers;
-	while (h) {
-		n = h->next;
-		g_free (h->name);
-		g_free (h->value);
-		g_free (h);
-		h = n;
+	header = *headers;
+	while (header) {
+		next = header->next;
+		g_free (header->name);
+		g_free (header->value);
+		
+		g_slice_free (HeaderRaw, header);
+		
+		header = next;
 	}
 	
 	*headers = NULL;
@@ -287,12 +302,16 @@ static void
 g_mime_parser_init (GMimeParser *parser, GMimeParserClass *klass)
 {
 	parser->priv = g_new (struct _GMimeParserPrivate, 1);
-	parser_init (parser, NULL);
-	
-	parser->priv->scan_from = FALSE;
-	parser->priv->have_regex = FALSE;
-	parser->priv->persist_stream = TRUE;
 	parser->priv->respect_content_length = FALSE;
+	parser->priv->persist_stream = TRUE;
+	parser->priv->have_regex = FALSE;
+	parser->priv->scan_from = FALSE;
+	
+#if defined (HAVE_GLIB_REGEX)
+	parser->priv->regex = NULL;
+#endif
+	
+	parser_init (parser, NULL);
 }
 
 static void
@@ -302,8 +321,13 @@ g_mime_parser_finalize (GObject *object)
 	
 	parser_close (parser);
 	
+#if defined (HAVE_GLIB_REGEX)
+	if (parser->priv->regex)
+		g_regex_unref (parser->priv->regex);
+#elif defined (HAVE_REGEX_H)
 	if (parser->priv->have_regex)
-		regfree (&parser->priv->header_regex);
+		regfree (&parser->priv->regex);
+#endif
 	
 	g_free (parser->priv);
 	
@@ -339,9 +363,18 @@ parser_init (GMimeParser *parser, GMimeStream *stream)
 	priv->headerleft = HEADER_INIT_SIZE - 1;
 	priv->headerptr = priv->headerbuf;
 	
-	priv->rawbuf = g_malloc (HEADER_RAW_INIT_SIZE);
-	priv->rawleft = HEADER_RAW_INIT_SIZE - 1;
-	priv->rawptr = priv->rawbuf;
+	if (offset == -1 || !priv->persist_stream) {
+		priv->rawbuf = g_malloc (HEADER_RAW_INIT_SIZE);
+		priv->rawleft = HEADER_RAW_INIT_SIZE - 1;
+		priv->rawptr = priv->rawbuf;
+	} else {
+		priv->rawbuf = NULL;
+		priv->rawptr = NULL;
+		priv->rawleft = 0;
+	}
+	
+	priv->message_headers_begin = -1;
+	priv->message_headers_end = -1;
 	
 	priv->headers_begin = -1;
 	priv->headers_end = -1;
@@ -481,9 +514,33 @@ g_mime_parser_get_persist_stream (GMimeParser *parser)
 void
 g_mime_parser_set_persist_stream (GMimeParser *parser, gboolean persist)
 {
+	struct _GMimeParserPrivate *priv;
+	
 	g_return_if_fail (GMIME_IS_PARSER (parser));
 	
-	parser->priv->persist_stream = persist;
+	priv = parser->priv;
+	
+	if (priv->persist_stream == persist)
+		return;
+	
+	if (persist) {
+		priv->persist_stream = TRUE;
+		
+		if (priv->seekable && !priv->rawbuf) {
+			priv->rawbuf = g_malloc (HEADER_RAW_INIT_SIZE);
+			priv->rawleft = HEADER_RAW_INIT_SIZE - 1;
+			priv->rawptr = priv->rawbuf;
+		}
+	} else {
+		priv->persist_stream = FALSE;
+		
+		if (priv->rawbuf) {
+			g_free (priv->rawbuf);
+			priv->rawbuf = NULL;
+			priv->rawptr = NULL;
+			priv->rawleft = 0;
+		}
+	}
 }
 
 
@@ -571,6 +628,9 @@ g_mime_parser_set_respect_content_length (GMimeParser *parser, gboolean respect_
  * Sets the regular expression pattern @regex on @parser. Whenever a
  * header matching the pattern @regex is parsed, @header_cb is called
  * with @user_data as the user_data argument.
+ *
+ * If @regex is %NULL, then the previously registered regex callback
+ * is unregistered and no new callback is set.
  **/
 void
 g_mime_parser_set_header_regex (GMimeParser *parser, const char *regex,
@@ -582,10 +642,17 @@ g_mime_parser_set_header_regex (GMimeParser *parser, const char *regex,
 	
 	priv = parser->priv;
 	
+#if defined (HAVE_GLIB_REGEX)
+	if (priv->regex) {
+		g_regex_unref (priv->regex);
+		priv->regex = NULL;
+	}
+#elif defined (HAVE_REGEX_H)
 	if (priv->have_regex) {
-		regfree (&priv->header_regex);
+		regfree (&priv->regex);
 		priv->have_regex = FALSE;
 	}
+#endif
 	
 	if (!regex || !header_cb)
 		return;
@@ -593,7 +660,11 @@ g_mime_parser_set_header_regex (GMimeParser *parser, const char *regex,
 	priv->header_cb = header_cb;
 	priv->user_data = user_data;
 	
-	priv->have_regex = !regcomp (&priv->header_regex, regex, REG_EXTENDED | REG_ICASE | REG_NOSUB);
+#if defined (HAVE_GLIB_REGEX)
+	priv->regex = g_regex_new (regex, G_REGEX_RAW | G_REGEX_EXTENDED | G_REGEX_CASELESS, 0, NULL);
+#elif defined (HAVE_REGEX_H)
+	priv->have_regex = !regcomp (&priv->regex, regex, REG_EXTENDED | REG_ICASE | REG_NOSUB);
+#endif
 }
 
 
@@ -642,7 +713,7 @@ parser_fill (GMimeParser *parser, size_t atleast)
 		priv->inend += nread;
 	}
 	
-	return priv->inend - priv->inptr;
+	return (ssize_t) (priv->inend - priv->inptr);
 }
 
 
@@ -732,12 +803,12 @@ parser_step_from (GMimeParser *parser)
 			
 			if (inptr + 1 >= inend) {
 				/* we don't have enough data; if we can't get more we have to bail */
-				left = inend - start;
+				left = (ssize_t) (inend - start);
 				priv->inptr = start;
 				goto refill;
 			}
 			
-			len = inptr - start;
+			len = (size_t) (inptr - start);
 			inptr++;
 			
 			if (len >= 5 && !strncmp (start, "From ", 5)) {
@@ -760,27 +831,49 @@ parser_step_from (GMimeParser *parser)
 	return 0;
 }
 
-static size_t
+#ifdef ALLOC_NEAREST_POW2
+static inline size_t
 nearest_pow (size_t num)
 {
-	size_t n = num > 0 ? num - 1 : 0;
+	size_t n;
 	
+	if (num == 0)
+		return 0;
+	
+	n = num - 1;
+#if defined (__GNUC__) && defined (__i386__)
+	__asm__("bsrl %1,%0\n\t"
+		"jnz 1f\n\t"
+		"movl $-1,%0\n"
+		"1:" : "=r" (n) : "rm" (n));
+	n = (1 << (n + 1));
+#else
 	n |= n >> 1;
 	n |= n >> 2;
 	n |= n >> 4;
 	n |= n >> 8;
 	n |= n >> 16;
 	n++;
+#endif
 	
 	return n;
 }
 
+#define next_alloc_size(n) nearest_pow (n)
+#else
+static inline size_t
+next_alloc_size (size_t n)
+{
+	return ((n + 63) / 64) * 64;
+}
+#endif
+
 #define header_append(priv, start, len) G_STMT_START {                    \
 	if (priv->headerleft <= len) {                                    \
 		size_t hlen, hoff;                                        \
 		                                                          \
 		hoff = priv->headerptr - priv->headerbuf;                 \
-		hlen = nearest_pow (hoff + len + 1);                      \
+		hlen = next_alloc_size (hoff + len + 1);                  \
 		                                                          \
 		priv->headerbuf = g_realloc (priv->headerbuf, hlen);      \
 		priv->headerptr = priv->headerbuf + hoff;                 \
@@ -793,33 +886,37 @@ nearest_pow (size_t num)
 } G_STMT_END
 
 #define raw_header_append(priv, start, len) G_STMT_START {                \
-	if (priv->rawleft <= len) {                                       \
-		size_t hlen, hoff;                                        \
+	if (priv->rawbuf) {                                               \
+		if (priv->rawleft <= len) {                               \
+			size_t hlen, hoff;                                \
+			                                                  \
+			hoff = priv->rawptr - priv->rawbuf;               \
+			hlen = next_alloc_size (hoff + len + 1);          \
+			                                                  \
+			priv->rawbuf = g_realloc (priv->rawbuf, hlen);    \
+			priv->rawptr = priv->rawbuf + hoff;               \
+			priv->rawleft = (hlen - 1) - hoff;                \
+		}                                                         \
 		                                                          \
-		hoff = priv->rawptr - priv->rawbuf;                       \
-		hlen = nearest_pow (hoff + len + 1);                      \
-		                                                          \
-		priv->rawbuf = g_realloc (priv->rawbuf, hlen);            \
-		priv->rawptr = priv->rawbuf + hoff;                       \
-		priv->rawleft = (hlen - 1) - hoff;                        \
+		memcpy (priv->rawptr, start, len);                        \
+		priv->rawptr += len;                                      \
+		priv->rawleft -= len;                                     \
 	}                                                                 \
-	                                                                  \
-	memcpy (priv->rawptr, start, len);                                \
-	priv->rawptr += len;                                              \
-	priv->rawleft -= len;                                             \
 } G_STMT_END
 
 #define raw_header_reset(priv) G_STMT_START {                             \
-	priv->rawleft += priv->rawptr - priv->rawbuf;                     \
-	priv->rawptr = priv->rawbuf;                                      \
+	if (priv->rawbuf) {                                               \
+		priv->rawleft += priv->rawptr - priv->rawbuf;             \
+		priv->rawptr = priv->rawbuf;                              \
+	}                                                                 \
 } G_STMT_END
 
 static void
-header_parse (GMimeParser *parser, struct _header_raw **tail)
+header_parse (GMimeParser *parser, HeaderRaw **tail)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
-	struct _header_raw *header;
 	register char *inptr, *end;
+	HeaderRaw *header;
 	char *start;
 	size_t hlen;
 	
@@ -842,7 +939,7 @@ header_parse (GMimeParser *parser, struct _header_raw **tail)
 	
 	hlen = inptr - priv->headerbuf;
 	
-	header = g_new (struct _header_raw, 1);
+	header = g_slice_new (HeaderRaw);
 	header->next = NULL;
 	
 	header->name = g_strndup (priv->headerbuf, hlen);
@@ -853,7 +950,7 @@ header_parse (GMimeParser *parser, struct _header_raw **tail)
 		inptr++;
 	
 	/* cut trailing lwsp */
-	start = inptr++;
+	start = inptr;
 	end = inptr;
 	
 	while (*inptr) {
@@ -861,7 +958,7 @@ header_parse (GMimeParser *parser, struct _header_raw **tail)
 			end = inptr;
 	}
 	
-	header->value = g_strndup (start, end - start);
+	header->value = g_strndup (start, (size_t) (end - start));
 	
 	header->offset = priv->header_offset;
 	
@@ -871,10 +968,16 @@ header_parse (GMimeParser *parser, struct _header_raw **tail)
 	priv->headerleft += priv->headerptr - priv->headerbuf;
 	priv->headerptr = priv->headerbuf;
 	
-	if (priv->have_regex &&
-	    !regexec (&priv->header_regex, header->name, 0, NULL, 0))
+#if defined (HAVE_GLIB_REGEX)
+	if (priv->regex && g_regex_match (priv->regex, header->name, 0, NULL))
 		priv->header_cb (parser, header->name, header->value,
 				 header->offset, priv->user_data);
+#elif defined (HAVE_REGEX_H)
+	if (priv->have_regex &&
+	    !regexec (&priv->regex, header->name, 0, NULL, 0))
+		priv->header_cb (parser, header->name, header->value,
+				 header->offset, priv->user_data);
+#endif
 }
 
 static int
@@ -883,20 +986,18 @@ parser_step_headers (GMimeParser *parser)
 	struct _GMimeParserPrivate *priv = parser->priv;
 	gboolean eoln, valid = TRUE, fieldname = TRUE;
 	gboolean continuation = FALSE;
-	struct _header_raw *hend;
 	register char *inptr;
 	char *start, *inend;
 	ssize_t left = 0;
+	HeaderRaw *tail;
 	size_t len;
 	
 	priv->midline = FALSE;
 	raw_header_reset (priv);
 	header_raw_clear (&priv->headers);
-	hend = (struct _header_raw *) &priv->headers;
-	priv->header_offset = parser_offset (priv, NULL);
-	
-	if (priv->headers_begin == -1)
-		priv->headers_begin = parser_offset (priv, NULL);
+	tail = (HeaderRaw *) &priv->headers;
+	priv->headers_begin = parser_offset (priv, NULL);
+	priv->header_offset = priv->headers_begin;
 	
 	inptr = priv->inptr;
 	inend = priv->inend;
@@ -918,7 +1019,7 @@ parser_step_headers (GMimeParser *parser)
 			
 			/* if we are scanning a new line, check for a folded header */
 			if (!priv->midline && continuation && (*inptr != ' ' && *inptr != '\t')) {
-				header_parse (parser, &hend);
+				header_parse (parser, &tail);
 				priv->header_offset = parser_offset (priv, inptr);
 				continuation = FALSE;
 				fieldname = TRUE;
@@ -943,7 +1044,7 @@ parser_step_headers (GMimeParser *parser)
 					
 					if (inptr == inend) {
 						/* don't have the full field name */
-						left = inend - start;
+						left = (ssize_t) (inend - start);
 						priv->inptr = start;
 						goto refill;
 					}
@@ -987,8 +1088,8 @@ parser_step_headers (GMimeParser *parser)
 				
 				raw_header_append (priv, start, len);
 				header_append (priv, start, len);
+				left = (ssize_t) (inend - inptr);
 				priv->midline = TRUE;
-				left = inend - inptr;
 				priv->inptr = inptr;
 				goto refill;
 			}
@@ -1011,7 +1112,7 @@ parser_step_headers (GMimeParser *parser)
 			inptr++;
 		}
 		
-		left = inend - inptr;
+		left = (ssize_t) (inend - inptr);
 		priv->inptr = inptr;
 	} while (1);
 	
@@ -1025,35 +1126,32 @@ parser_step_headers (GMimeParser *parser)
  headers_end:
 	
 	if (priv->headerptr > priv->headerbuf)
-		header_parse (parser, &hend);
-	
-	if (priv->headers_end == -1)
-		priv->headers_end = parser_offset (priv, start);
+		header_parse (parser, &tail);
 	
+	priv->headers_end = parser_offset (priv, start);
 	priv->state = GMIME_PARSER_STATE_HEADERS_END;
-	*priv->rawptr = '\0';
+	if (priv->rawbuf)
+		*priv->rawptr = '\0';
 	priv->inptr = inptr;
 	
 	return 0;
 	
  next_message:
 	
-	if (priv->headers_end == -1)
-		priv->headers_end = parser_offset (priv, start);
-	
+	priv->headers_end = parser_offset (priv, start);
 	priv->state = GMIME_PARSER_STATE_COMPLETE;
-	*priv->rawptr = '\0';
+	if (priv->rawbuf)
+		*priv->rawptr = '\0';
 	priv->inptr = start;
 	
 	return 0;
 	
  content_start:
 	
-	if (priv->headers_end == -1)
-		priv->headers_end = parser_offset (priv, start);
-	
+	priv->headers_end = parser_offset (priv, start);
 	priv->state = GMIME_PARSER_STATE_CONTENT;
-	*priv->rawptr = '\0';
+	if (priv->rawbuf)
+		*priv->rawptr = '\0';
 	priv->inptr = start;
 	
 	return 0;
@@ -1064,7 +1162,8 @@ content_type_destroy (ContentType *content_type)
 {
 	g_free (content_type->subtype);
 	g_free (content_type->type);
-	g_free (content_type);
+	
+	g_slice_free (ContentType, content_type);
 }
 
 static gboolean
@@ -1087,11 +1186,10 @@ static ContentType *
 parser_content_type (GMimeParser *parser)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
-	register const char *inptr;
 	ContentType *content_type;
 	const char *value;
 	
-	content_type = g_new (ContentType, 1);
+	content_type = g_slice_new (ContentType);
 	
 	if (!(value = header_raw_find (priv->headers, "Content-Type", NULL)) ||
 	    !g_mime_parse_content_type (&value, &content_type->type, &content_type->subtype)) {
@@ -1148,20 +1246,25 @@ parser_step (GMimeParser *parser)
 	case GMIME_PARSER_STATE_ERROR:
 		break;
 	case GMIME_PARSER_STATE_INIT:
-		parser->priv->headers_begin = -1;
-		parser->priv->headers_end = -1;
+		priv->message_headers_begin = -1;
+		priv->message_headers_end = -1;
 		if (priv->scan_from)
 			priv->state = GMIME_PARSER_STATE_FROM;
 		else
 			priv->state = GMIME_PARSER_STATE_HEADERS;
 		break;
 	case GMIME_PARSER_STATE_FROM:
-		parser->priv->headers_begin = -1;
-		parser->priv->headers_end = -1;
+		priv->message_headers_begin = -1;
+		priv->message_headers_end = -1;
 		parser_step_from (parser);
 		break;
 	case GMIME_PARSER_STATE_HEADERS:
 		parser_step_headers (parser);
+		
+		if (priv->message_headers_begin == -1) {
+			priv->message_headers_begin = priv->headers_begin;
+			priv->message_headers_end = priv->headers_end;
+		}
 		break;
 	case GMIME_PARSER_STATE_HEADERS_END:
 		if (parser_skip_line (parser) == -1)
@@ -1203,11 +1306,11 @@ check_boundary (struct _GMimeParserPrivate *priv, const char *start, size_t len)
 {
 	gint64 offset = parser_offset (priv, start);
 	
-	if (start[len - 1] == '\r')
+	if (len > 0 && start[len - 1] == '\r')
 		len--;
 	
 	if (possible_boundary (priv->scan_from, start, len)) {
-		struct _boundary_stack *s;
+		BoundaryStack *s;
 		
 		d(printf ("checking boundary '%.*s'\n", len, start));
 		
@@ -1254,7 +1357,7 @@ check_boundary (struct _GMimeParserPrivate *priv, const char *start, size_t len)
 #define MAX_BOUNDARY_LEN(bounds) (bounds ? bounds->boundarylenmax + 2 : 0)
 
 static int
-parser_scan_content (GMimeParser *parser, GByteArray *content, int *crlf)
+parser_scan_content (GMimeParser *parser, GByteArray *content, guint *crlf)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
 	register char *inptr;
@@ -1356,7 +1459,7 @@ parser_scan_mime_part_content (GMimeParser *parser, GMimePart *mime_part, int *f
 	GMimeDataWrapper *wrapper;
 	GMimeStream *stream;
 	gint64 start, end;
-	int crlf;
+	guint crlf;
 	
 	g_assert (priv->state >= GMIME_PARSER_STATE_HEADERS_END);
 	
@@ -1395,10 +1498,11 @@ static void
 parser_scan_message_part (GMimeParser *parser, GMimeMessagePart *mpart, int *found)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
-	struct _header_raw *header;
 	ContentType *content_type;
 	GMimeMessage *message;
 	GMimeObject *object;
+	GMimeStream *stream;
+	HeaderRaw *header;
 	
 	g_assert (priv->state == GMIME_PARSER_STATE_CONTENT);
 	
@@ -1427,6 +1531,10 @@ parser_scan_message_part (GMimeParser *parser, GMimeMessagePart *mpart, int *fou
 	content_type_destroy (content_type);
 	message->mime_part = object;
 	
+	/* set the same raw header stream on the message's header-list */
+	if ((stream = g_mime_header_list_get_stream (object->headers)))
+		g_mime_header_list_set_stream (((GMimeObject *) message)->headers, stream);
+	
 	g_mime_message_part_set_message (mpart, message);
 	g_object_unref (message);
 }
@@ -1435,8 +1543,9 @@ static GMimeObject *
 parser_construct_leaf_part (GMimeParser *parser, ContentType *content_type, int *found)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
-	struct _header_raw *header;
 	GMimeObject *object;
+	GMimeStream *stream;
+	HeaderRaw *header;
 	
 	g_assert (priv->state >= GMIME_PARSER_STATE_HEADERS_END);
 	
@@ -1458,7 +1567,15 @@ parser_construct_leaf_part (GMimeParser *parser, ContentType *content_type, int
 	
 	header_raw_clear (&priv->headers);
 	
-	g_mime_header_list_set_raw (object->headers, priv->rawbuf);
+	/* set the raw header stream on the header-list */
+	if (priv->persist_stream && priv->seekable)
+		stream = g_mime_stream_substream (priv->stream, priv->headers_begin, priv->headers_end);
+	else
+		stream = g_mime_stream_mem_new_with_buffer (priv->rawbuf, priv->rawptr - priv->rawbuf);
+	
+	g_mime_header_list_set_stream (object->headers, stream);
+	g_object_unref (stream);
+	
 	raw_header_reset (priv);
 	
 	if (priv->state == GMIME_PARSER_STATE_HEADERS_END) {
@@ -1506,8 +1623,9 @@ static int
 parser_scan_multipart_face (GMimeParser *parser, GMimeMultipart *multipart, gboolean preface)
 {
 	GByteArray *buffer;
-	int found, crlf;
 	char *face;
+	guint crlf;
+	int found;
 	
 	buffer = g_byte_array_new ();
 	found = parser_scan_content (parser, buffer, &crlf);
@@ -1536,7 +1654,7 @@ parser_scan_multipart_face (GMimeParser *parser, GMimeMultipart *multipart, gboo
 static gboolean
 found_immediate_boundary (struct _GMimeParserPrivate *priv, gboolean end)
 {
-	struct _boundary_stack *s = priv->bounds;
+	BoundaryStack *s = priv->bounds;
 	size_t len = end ? s->boundarylenfinal : s->boundarylen;
 	
 	return !strncmp (priv->inptr, s->boundary, len)
@@ -1588,10 +1706,11 @@ static GMimeObject *
 parser_construct_multipart (GMimeParser *parser, ContentType *content_type, int *found)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
-	struct _header_raw *header;
 	GMimeMultipart *multipart;
 	const char *boundary;
 	GMimeObject *object;
+	GMimeStream *stream;
+	HeaderRaw *header;
 	
 	g_assert (priv->state >= GMIME_PARSER_STATE_HEADERS_END);
 	
@@ -1605,7 +1724,15 @@ parser_construct_multipart (GMimeParser *parser, ContentType *content_type, int
 	
 	header_raw_clear (&priv->headers);
 	
-	g_mime_header_list_set_raw (object->headers, priv->rawbuf);
+	/* set the raw header stream on the header-list */
+	if (priv->persist_stream && priv->seekable)
+		stream = g_mime_stream_substream (priv->stream, priv->headers_begin, priv->headers_end);
+	else
+		stream = g_mime_stream_mem_new_with_buffer (priv->rawbuf, priv->rawptr - priv->rawbuf);
+	
+	g_mime_header_list_set_stream (object->headers, stream);
+	g_object_unref (stream);
+	
 	raw_header_reset (priv);
 	
 	multipart = (GMimeMultipart *) object;
@@ -1692,10 +1819,11 @@ parser_construct_message (GMimeParser *parser)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
 	unsigned long content_length = ULONG_MAX;
-	struct _header_raw *header;
 	ContentType *content_type;
 	GMimeMessage *message;
 	GMimeObject *object;
+	GMimeStream *stream;
+	HeaderRaw *header;
 	char *endptr;
 	int found;
 	
@@ -1739,6 +1867,10 @@ parser_construct_message (GMimeParser *parser)
 	content_type_destroy (content_type);
 	message->mime_part = object;
 	
+	/* set the same raw header stream on the message's header-list */
+	if ((stream = g_mime_header_list_get_stream (object->headers)))
+		g_mime_header_list_set_stream (((GMimeObject *) message)->headers, stream);
+	
 	if (priv->scan_from) {
 		priv->state = GMIME_PARSER_STATE_FROM;
 		parser_pop_boundary (parser);
@@ -1833,7 +1965,7 @@ g_mime_parser_get_headers_begin (GMimeParser *parser)
 {
 	g_return_val_if_fail (GMIME_IS_PARSER (parser), -1);
 	
-	return parser->priv->headers_begin;
+	return parser->priv->message_headers_begin;
 }
 
 
@@ -1852,5 +1984,5 @@ g_mime_parser_get_headers_end (GMimeParser *parser)
 {
 	g_return_val_if_fail (GMIME_IS_PARSER (parser), -1);
 	
-	return parser->priv->headers_end;
+	return parser->priv->message_headers_end;
 }
